package edu.stanford.nlp.tagger.maxent;

import java.util.*;
import java.io.*;

public class ASBCDict {
  private final static String defaultFilename = "adbcdict";
  private static String pathPrefix;
  private static ASBCDict ASBCDictSingleton = null;

  public static ASBCDict getInstance()  {

     if (ASBCDictSingleton == null) {
        ASBCDictSingleton = new ASBCDict();
     }
     return ASBCDictSingleton;
  }


  private ASBCDict() {
    readASBCDict("/u/nlp/data/pos-tagger/dictionary" + "/" + defaultFilename);
  }


  public static HashMap <String, Set <String>> ASBC_pre_dict;
  public static HashMap <String, Set <String>> ASBC_suf_dict;


  private static void readASBCDict(String filename)   {
    try {

      BufferedReader ASBCDetectorReader = new BufferedReader(new InputStreamReader(new FileInputStream(filename), "GB18030"));
      String ASBCDetectorLine;

      ASBC_pre_dict = new HashMap<String, Set <String>>();
      ASBC_suf_dict = new HashMap<String, Set <String>>();

      while ((ASBCDetectorLine = ASBCDetectorReader.readLine()) != null) {
        String[] fields = ASBCDetectorLine.split("	");
        String tag=fields[0];
        Set<String> pres=ASBC_pre_dict.get(tag);
        Set<String> sufs=ASBC_suf_dict.get(tag);
        //System.err.println(fields[0]+fields[1]+fields[2]);
        if (pres==null) {
          pres = new HashSet<String>();
          ASBC_pre_dict.put(tag,pres);
        }
        pres.add(fields[1]);

        if (sufs==null) {
          sufs = new HashSet<String>();
          ASBC_suf_dict.put(tag,sufs);
        }
        sufs.add(fields[2]);
      }
      ASBCDetectorReader.close();
    } catch (FileNotFoundException e) {
      System.err.println("ASBC not found");
      System.exit(-1);
    } catch (IOException e) {
      System.err.println("ASBC");
      System.exit(-1);
    }
  }


 public static String getTagPre(String a1, String a2) {
   ASBCDict dict = ASBCDict.getInstance();
    if (dict.getpre(a1)== null)	return "0";
    if (dict.getpre(a1).contains(a2))
      return "1";
    return "0";
  }



 public static String getTagSuf(String a1, String a2) {
   ASBCDict dict = ASBCDict.getInstance();
    if (dict.getsuf(a1)== null)	return "0";
    if (dict.getsuf(a1).contains(a2))
      return "1";
    return "0";
  }


  Set getpre(String a) {
    return ASBC_pre_dict.get(a);
  }

  Set getsuf(String a){
    return ASBC_suf_dict.get(a);
  }


  public static String getPathPrefix() {
    return pathPrefix;
  }


  public static void setPathPrefix(String pathPrefix) {
    ASBCDict.pathPrefix = pathPrefix;
  }

} //class
