/**
 * Title:        StanfordMaxEnt<p>
 * Description:  A Maximum Entropy Toolkit<p>
 * Copyright:    Copyright (c) Kristina Toutanova<p>
 * Company:      Stanford University<p>
 * @author Kristina Toutanova
 * @version 1.0
 */
package edu.stanford.nlp.tagger.maxent;

import java.util.HashMap;
import java.util.Map;

import edu.stanford.nlp.util.FileLines;
import edu.stanford.nlp.util.Timing;

public class ExtractorDistsim extends Extractor {

  private static final long serialVersionUID = 1L;

  
  private static String defaultDistsimPath = "/u/nlp/data/pos_tags_are_useless/egw.bnc.pruned.200";
    
  
  private Map<String,String> lexicon;

  private void initLexicon(String path) {
    Timing.startDoing("Loading distsim lexicon from " + path);
    lexicon = new HashMap<String, String>();
    for (String word : new FileLines(path)) {
      String[] bits = word.split("\\s+");
      lexicon.put(bits[0].toLowerCase(), bits[1]);
    }
    Timing.endDoing();
  }

  public boolean Precondition(String tag) {
    return true;
  }


  String extract(History h) {
    String curWord = ExtractorFrames.cWord.extract(h);
    String distSim = lexicon.get(curWord.toLowerCase());
    if(distSim == null) distSim = "null";
    return distSim;
  }

  public ExtractorDistsim(String distSimPath) {
    initLexicon(distSimPath);
  }

}
