/**
 * Title:        StanfordMaxEnt<p>
 * Description:  A Maximum Entropy Toolkit<p>
 * Copyright:    Copyright (c) Kristina Toutanova<p>
 * Company:      Stanford University<p>
 * @author Kristina Toutanova
 * @version 1.0
 */
package edu.stanford.nlp.tagger.maxent;

public class ExtractorLastVerb extends Extractor {

  private static final String partTag = "RP";
  private static final String inTag = "IN";
  private static final String rbTag = "RB";
  private static final String rareVerb = "rareverbed";
  private final int bound;

  public ExtractorLastVerb() {
    this(8);
  }

  public ExtractorLastVerb(int bound) {
    this.bound = bound;
  }


  public boolean Precondition(String tag) {
    return (tag.equals(partTag) || (tag.equals(inTag)) || (tag.equals(rbTag)));
  }


  String extract(History h) {
    // should extract last verbal word and also the current word
    String cword = ExtractorFrames.cWord.extract(h);
    int allCount = GlobalHolder.dict.sum(cword);
    int rpCount = GlobalHolder.dict.getCount(cword, "RP");
    int rbCount = GlobalHolder.dict.getCount(cword, "RB");
    int inCount = GlobalHolder.dict.getCount(cword, "IN");
    if ((allCount == 0) || ((allCount > 0) && (rpCount + rbCount + inCount <= allCount / 100))) {
      return zeroSt;
    }
    String lastverb = extractLV(h, bound);
    if (lastverb.startsWith("NA")) {
      return "2";
    }
    if (GlobalHolder.dict.sum(lastverb) < 2) {
      return rareVerb + "|" + cword;
    }
    return lastverb + "|" + cword;
  }


  String extractLV(History h) {
    // should extract last verbal word and also the current word
    int start = h.start;
    String lastverb = "NA";
    int current = h.current;
    int index = current - 1;
    while (index >= start) {
      String tag = GlobalHolder.pairs.get(index, true);
      if (tag.startsWith("VB")) {
        lastverb = GlobalHolder.pairs.get(index, false);
        break;
      }
      if (tag.startsWith(",")) {
        break;
      }
      index--;
    }
    return lastverb;
  }

  String extractLV(History h, int bound) {
    // should extract last verbal word and also the current word
    int start = h.start;
    String lastverb = "NA";
    int current = h.current;
    int index = current - 1;
    while ((index >= start) && (index >= current - bound)) {
      String tag = GlobalHolder.pairs.get(index, true);
      if (tag.startsWith("VB")) {
        lastverb = GlobalHolder.pairs.get(index, false);
        break;
      }
      if (tag.startsWith(",")) {
        break;
      }
      index--;
    }
    return lastverb;
  }

  private static final long serialVersionUID = -16619310062224383L;

} 
