/**
 * Title:        StanfordMaxEnt<p>
 * Description:  A Maximum Entropy Toolkit<p>
 * Copyright:    Copyright (c) Kristina Toutanova<p>
 * Company:      Stanford University<p>
 */

package edu.stanford.nlp.tagger.maxent;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import edu.stanford.nlp.io.PrintFile;
import edu.stanford.nlp.ling.HasWord;
import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.ling.WordTag;
import edu.stanford.nlp.maxent.iis.LambdaSolve;
import edu.stanford.nlp.sequences.BestSequenceFinder;
import edu.stanford.nlp.sequences.ExactBestSequenceFinder;
import edu.stanford.nlp.sequences.SequenceModel;

/**
 * @author Kristina Toutanova
 * @version 1.0
 */
public class TestSentence implements SequenceModel {
  public static String delimiter = "/";
  public LambdaSolve prob;
  public static int[] hPos;
  public static boolean[] isTag;
  private String sentence;
  private List<String> sent = new ArrayList<String>();
  static boolean VERBOSE = false;
  static final boolean DBG = false;
  private int size;
  private int beamSize = 10;
  private String[][] nBest;
  private double[] pBest;
  private String[] correctTags;
  private String[] tags;
  public int numRight;
  public int numWrong;
  int numUnknown;
  int numWrongUnknown;
  int numUnassigned, numUnknownUnassigned;
  int numWrongUnknownT3;
  int numWrongT3;
  private int startSizePairs = GlobalHolder.pairs.getSize();
  private int endSizePairs = startSizePairs;
  static final String naTag = "NA";
  static final String eosWord = "EOS";
  private double[][][] probabilities;
  private static final double sigma = 1.1;
  private boolean taginference = true;
  static int leftContext = 2;
  static int rightContext = 2;
  private static boolean doDeterministicTagExpansion = true;


  public TestSentence() {
  }

  public TestSentence(LambdaSolve prob, String s) {
    this(prob, s, null);
  }

  public TestSentence(LambdaSolve prob, String s, PrintFile pf) {
    this.prob = prob;
    this.sentence = s;
    init();
    test();
    if(pf != null) {
      pf.print(this.getTaggedNice());
      pf.print("\n");
    }
    revert(startSizePairs, endSizePairs);
  }

  public TestSentence(LambdaSolve prob, Sentence s, PrintFile pf) {
    this.prob = prob;
    this.sentence = s.toString(true);

    for(Object word : s) {
      sent.add(word.toString());
    }
    init2();
    test();
  }



  public TestSentence(LambdaSolve prob, String[] s, String[] correctTags, PrintFile pf, Dictionary wrongWords) {

    this.prob = prob;
    this.sent = new ArrayList<String>();
    for(int i = 0; i < s.length; i++) {
      this.sent.add(s[i]);
    }
    this.correctTags = correctTags;
    this.tags = new String[correctTags.length];
    for (int i = 0; i < correctTags.length; i++) {
      tags[i] = naTag;
    }
    init1();

    if (!taginference) {
      test1(pf, wrongWords);
      revert(startSizePairs, endSizePairs);
    } else {
      testTagInference(pf, wrongWords);
    }

  }


  public TestSentence(LambdaSolve prob, String[] s, String[] tags, String[] correctTags, PrintFile pf, Dictionary wrongWords) {

    this.prob = prob;
    this.sent = new ArrayList<String>();
    for(int i = 0; i < s.length; i++) {
      this.sent.add(s[i]);
    }
    this.correctTags = correctTags;
    this.tags = tags;
    init1();
    test1(pf, wrongWords);
    revert(startSizePairs, endSizePairs);
  }


  /**
   * Tags the sentence s by solving prob.  Returns a sentence of
   * TaggedWord objects.
   * @param prob
   * @param s
   * @return
   */
  public Sentence<TaggedWord> tagSentence(LambdaSolve prob, Sentence<? extends HasWord> s) {
    this.prob = prob;
    this.sent = new ArrayList<String>();//new String[s.length() + 1];
    for (int j = 0; j < s.length(); j++) {
      //this.sent[j] = s.getWord(j).value();
      this.sent.add(((HasWord)s.get(j)).word());
    }
    size = sent.size();
    sent.add(eosWord);
    //this.sentence=s;
    if (VERBOSE) {
      System.out.println("Sentence is " + s.toString());
    }
    init1();
    tags = new String[size];
    for (int j = 0; j < size; j++) {
      tags[j] = naTag;
    }
    test();
    revert(startSizePairs, endSizePairs);

    ArrayList<TaggedWord> taggedWords = new ArrayList<TaggedWord>();
    for (int j = 0; j < sent.size() - 1; j++) {
      String tag = nBest[0][j];
      TaggedWord w = new TaggedWord(sent.get(j), tag);
      taggedWords.add(w);
    }

    return new Sentence<TaggedWord>(taggedWords);
  }



  public void revert(int prevSize, int afterSize) {
    GlobalHolder.pairs.remove(prevSize, afterSize);
    endSizePairs = prevSize;
    CollectionTaggerOutputs.baseToken += (afterSize - prevSize);
  }


  public void init() {
    StringTokenizer sT = new StringTokenizer(sentence);
    size = 0;
    while (sT.hasMoreTokens()) {
      String word = sT.nextToken();
      sent.add(word);
      size++;
    }

    sent.add(eosWord);
    size++;
    nBest = new String[beamSize][size];
    pBest = new double[beamSize];
    tags = new String[size];
    for (int i = 0; i < size; i++) {
      tags[i] = naTag;
    }
  }

  public void init1() {

    //the eos are assumed already there
    size = sent.size();

    nBest = new String[beamSize][size];
    pBest = new double[beamSize];

    for (int i = 0; i < size - 1; i++) {
      if (GlobalHolder.dict.sum(sent.get(i)) == 0) {
        numUnknown++;
      }

    }
  }

  public void init2() {

    //the eos are assumed already there
    size = sent.size();

    nBest = new String[beamSize][size];
    pBest = new double[beamSize];

    tags = new String[size];
    for (int i = 0; i < size; i++) {
      tags[i] = naTag;
    }
  }


  /**
   * Returns a string representation of the sentence.
   * @return
   */
  public String getTaggedNice() {
    StringBuilder sb = new StringBuilder();
    for (int i = 0; i < size - 1; i++) {
      sb.append(toNice(sent.get(i))).append(delimiter).append(toNice(nBest[0][i]));
      sb.append(" ");
    }
    return sb.toString();
  }

  public Sentence<TaggedWord> getTaggedSentence() {
    Sentence<TaggedWord> taggedSentence = new Sentence<TaggedWord>();

    for (int j = 0; j < sent.size() - 1; j++) {
      String tag = nBest[0][j];
      TaggedWord w = new TaggedWord(sent.get(j), tag);
      taggedSentence.add(w);
    }
    return taggedSentence;
  }

  public static String toNice(String s) {
    if (s == null) {
      return " NA ";
    } else {
      return s;
    }

  }


  /**
   * tag a sentence using Dan's TagScorer interface
   */

  public static void tagSentenceTagScorer(List<String> sent) {
    TestSentence ts = new TestSentence();
    ts.initializeScorer(GlobalHolder.getLambdaSolve(), sent);
    BestSequenceFinder ti = new ExactBestSequenceFinder();//new BeamBestSequenceFinder(15);
    int[] bestTags = ti.bestSequence(ts);
    for (int j = 0; j < sent.size(); j++) {
      System.out.print(sent.get(j) + delimiter + GlobalHolder.tags.getTag(bestTags[j + ts.leftWindow()]) + "\t");
    }

    ts.cleanUpScorer();

  }





  public void writeProbs() {
    // for the hypotheses equal 1 to 5
    int numTags = GlobalHolder.tags.getSize();
    probabilities = new double[size][beamSize][numTags];
    double[] probs;
    for (int hyp = 0; hyp < beamSize; hyp++) {
      // put the whole thing on the pairs, give its beginning and end
      for (int i = 0; i < size; i++) {
        WordTag wT = new WordTag(sent.get(i), nBest[hyp][i]);
        GlobalHolder.pairs.add(wT);
      }
      int start = endSizePairs;
      int end = endSizePairs + size - 1;
      endSizePairs = endSizePairs + size;
      // iterate over the sentence
      for (int current = 0; current < size; current++) {
        History h = new History(start, end, current + start);
        probs = getHistories(h);
        double s = 0;
        for (int k = 0; k < numTags; k++) {
          s += probs[k];
        }
        for (int k = 0; k < numTags; k++) {
          probabilities[current][hyp][k] = probs[k] / s;
        }

      } // for current

    } // for hyp

  }// end


  /**
   * tokenize s into words, and dump unknown word activations
   */
  public void dumpActivations(String s) {

    this.sentence = s;
    init();

    for (int i = 0; i < size; i++) {
      WordTag wT = new WordTag(sent.get(i), naTag);
      GlobalHolder.pairs.add(wT);
    }

    int start = endSizePairs;
    int end = endSizePairs + size - 1;
    endSizePairs = endSizePairs + size;
    // iterate over the sentence
    for (int current = 0; current < size; current++) {
      History h = new History(start, end, current + start);
      printActivations(h);

    } // for current

    revert(startSizePairs, endSizePairs);
  }


  public String[] test() {
    int current = 0;
    for (int j = 0; j < beamSize; j++) {
      for (int i = 0; i < size; i++) {
        nBest[j][i] = naTag;
        pBest[j] = 1;
      }
    }

    while (current < size) {
      if (VERBOSE) {
        System.out.println("current is " + current + " word " + sent.get(current));
      }
      String[][] hCurrent = new String[beamSize][size];
      double[] pCurrent = new double[beamSize];
      for (int j = 0; j < pCurrent.length; j++) {
        pCurrent[j] = Double.NEGATIVE_INFINITY;
      } // * remove
      insertTags(current, hCurrent, pCurrent);
      if (VERBOSE) {
        System.out.println("current is " + current + " word " + sent.get(current));
      }

      for (int i = 0; i < beamSize; i++) {
        if (hCurrent[i][0] == null) {
          // fill it and to the end with hCurrent[i-1]
          for (int s = i; s < beamSize; s++) {
            for (int p = 0; p < current + 1; p++) {
              hCurrent[s][p] = hCurrent[i - 1][p];
            }
            pCurrent[s] = pCurrent[i - 1];
          }
          break;
        }  // end if
      }
      nBest = hCurrent;
      pBest = pCurrent;
      current++;
    }
//  for(int i = 0; i < size -1; i++) {
//  System.out.print(sent.get(i) + "/" + nBest[0][i] + " ");
//  }
//  System.out.println();
    return nBest[0];
  }


  public void test1(PrintFile pf, Dictionary wrongWords) {
    int current = 0;
    for (int j = 0; j < beamSize; j++) {
      for (int i = 0; i < size; i++) {
        nBest[j][i] = naTag;
        pBest[j] = 0;  //* change back to 1
      }
    }

    while (current < size) {
      String[][] hCurrent = new String[beamSize][size];
      double[] pCurrent = new double[beamSize];
      for (int j = 0; j < pCurrent.length; j++) {
        pCurrent[j] = Double.NEGATIVE_INFINITY;
      } // * remove
      insertTags(current, hCurrent, pCurrent);
      for (int i = 0; i < beamSize; i++) {
        if (hCurrent[i][0] == null) {
          // fill it and to the end with hCurrent[i-1]
          for (int s = i; s < beamSize; s++) {
            for (int p = 0; p < current + 1; p++) {
              hCurrent[s][p] = hCurrent[i - 1][p];
            }
            pCurrent[s] = pCurrent[i - 1];
          }
          break;
        }  // end if
      }
      nBest = hCurrent;
      pBest = pCurrent;
      current++;
    }
    GlobalHolder.collectionTaggers.baseToken += size;

    int finalHyp = 0;

    try {
      //without the EOS word
      for (int i = 0; i < correctTags.length - 1; i++) {

        if(pf != null) pf.print(toNice(sent.get(i)) + '_' + nBest[finalHyp][i]);
        if ((correctTags[i]).equals(nBest[finalHyp][i])) {
          numRight++;
        } else {
          numWrong++;
          System.out.println("Word: " + sent.get(i) + "; correct: " + correctTags[i] + "; guessed: " + nBest[finalHyp][i]);
          if(pf != null) pf.print("|" + correctTags[i]);
          wrongWords.add(sent.get(i) + correctTags[i], nBest[finalHyp][i]);
          if (GlobalHolder.dict.sum(sent.get(i)) == 0) {
            numWrongUnknown++;
            if(pf != null) pf.print("*");
          } // if
        } // else
        if(pf != null) pf.print(" ");
      }// for
      if(pf != null) pf.println();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }


  /**
   * Test using TagInference.
   */
  public void testTagInference(PrintFile pf, Dictionary wrongWords) {
    this.initializeScorer(GlobalHolder.getLambdaSolve(), sent);

    BestSequenceFinder ti = new ExactBestSequenceFinder();//new BeamBestSequenceFinder(50);//
    int[] bestTags = ti.bestSequence(this);
    String[] finalTags = new String[bestTags.length];
    for (int j = 0; j < sent.size(); j++) {
      finalTags[j] = GlobalHolder.tags.getTag(bestTags[j + leftWindow()]);
    }

    cleanUpScorer();

    try {

      for (int i = 0; i < correctTags.length - 1; i++) {
        boolean nReliable = false;
        if (true) {
          nReliable = true;
        }
        if ((correctTags[i]).equals(finalTags[i])) {
          if(pf != null) pf.print(toNice(sent.get(i)) + "_" + finalTags[i] + " ");
          numRight++;
        } else {
          numWrong++;
          System.out.println("Word: " + sent.get(i) + "; correct: " + correctTags[i] + "; guessed: " + finalTags[i]);

          if (nReliable) {
            numUnassigned++;
          }
          if(pf != null) pf.print(toNice(sent.get(i)) + "_" + finalTags[i] + "|" + correctTags[i]);
          wrongWords.add(sent.get(i) + correctTags[i], finalTags[i]);
          if (GlobalHolder.dict.sum(sent.get(i)) == 0) {
            numWrongUnknown++;
            if(pf != null) pf.print("*");
            if (nReliable) {
              numUnknownUnassigned++;
            }
          }// if
          if(pf != null) pf.print(" ");
        }// else
      }// for
      if(pf != null) pf.println();

    } catch (Exception e) {
      e.printStackTrace();
    }
  }


  /**
   * Test using TagInference.
   */
  public Sentence tagInference(Sentence s) {
    this.initializeScorer(GlobalHolder.getLambdaSolve(), sent);

    BestSequenceFinder ti = new ExactBestSequenceFinder();
    int[] bestTags = ti.bestSequence(this);
    String[] finalTags = new String[bestTags.length];
    for (int j = 0; j < sent.size(); j++) {
      finalTags[j] = GlobalHolder.tags.getTag(bestTags[j + leftWindow()]);
    }

    cleanUpScorer();

    ArrayList<TaggedWord> taggedWords = new ArrayList<TaggedWord>();
    for (int j = 0; j < sent.size() - 1; j++) {
      String tag = finalTags[j];//nBest[0][j];
      TaggedWord w = new TaggedWord(sent.get(j), tag);
      taggedWords.add(w);
    }

    return new Sentence(taggedWords);
  }

  public boolean known(String w) {
    return GlobalHolder.dict.sum(w) > 0;
  }


  public boolean reliable(int current) {
    String tag = nBest[0][current];
    int y = GlobalHolder.tags.getIndex(tag);
    double max = 0;
    int maxInd = -1;
    double p = probabilities[current][0][y];
    if (known(sent.get(current))) {
      System.out.println(" known " + sent.get(current));
      String[] tags = GlobalHolder.dict.getTags(sent.get(current));
      String[] tags1 = append(tags, sent.get(current));
      for (int k = 0; k < tags1.length; k++) {
        int i = GlobalHolder.tags.getIndex(tags1[k]);
        if (i == y) {
          continue;
        }
        if ((probabilities[current][0][i] >= max) && (probabilities[current][0][i] <= p)) {
          max = probabilities[current][0][i];
          maxInd = i;
        }
      }
    }// if known
    else {
      System.out.println(" unknown " + sent.get(current));
      for (int i = 0; i < GlobalHolder.ySize; i++) {
        if (i == y) {
          continue;
        }
        if ((probabilities[current][0][i] >= max) && (probabilities[current][0][i] <= p)) {
          max = probabilities[current][0][i];
          maxInd = i;
        }
      }
    }// else
    if (maxInd == -1) {
      return true;
    }
    String tag1 = GlobalHolder.tags.getTag(maxInd);
    System.out.println(tag + " " + GlobalHolder.tags.getTag(maxInd) + " " + p / max);
    if ((GlobalHolder.dict.sum(sent.get(current)) == 0) && (GlobalHolder.tags.isClosed(tag1))) {
      return true;
    } else {
      return (p / max) > sigma;
    }
  }


  /**
   * this is used for Dan's tag inference methods
   * current is the actual word number + leftW
   */
  void setHistory(int current, History h, int[] tags) {
    //writes over the tags in the last thing in pairs

    int left = leftWindow();
    int right = rightWindow();

    for (int j = current - left; j <= current + right; j++) {
      if (j < left) {
        continue;
      } //but shouldn't happen
      if (j >= size + left) {
        break;
      } //but shouldn't happen
      h.setTag(j - left, GlobalHolder.tags.getTag(tags[j]));
    }

  }

  /**
   * do initializations for the TagScorer interface
   */
  void initializeScorer(LambdaSolve prob, List<String> sentence) {
    this.sent = sentence;
    this.prob = prob;
    size = sent.size();
    for (int i = 0; i < size; i++) {
      WordTag wT = new WordTag(sent.get(i), "NN");
      GlobalHolder.pairs.add(wT);

    }

    endSizePairs += size;
  }


  /**
   * clean-up after the scorer
   */
  void cleanUpScorer() {
    revert(startSizePairs, endSizePairs);
  }


  void getHistory(int current, History h, int hyp) {

    for (int i = 0; i < current; i++) {
      WordTag wT = new WordTag(sent.get(i), nBest[hyp][i]);
      GlobalHolder.pairs.add(wT);
    }
    for (int i = current; i < size; i++) {

      WordTag wT = new WordTag(sent.get(i), tags[i]);
      GlobalHolder.pairs.add(wT);

    }

    h.set(endSizePairs, endSizePairs + size - 1, current + endSizePairs);
    this.endSizePairs = endSizePairs + size;

  }


  void insertTags(int current, String[][] hCurrent, double[] pCurrent) {
    History h = new History();
    double sum = 0.0;
    int x;
    boolean hasHistory = true;
    double[] histories = null;

    int allCount = GlobalHolder.dict.sum(sent.get(current));
    if (allCount == 0) {
    }
    //numUnknown++;
    for (int hyp = 0; hyp < beamSize; hyp++) {
      // Find the history
      try {
        getHistory(current, h, hyp);
      } catch (Exception e) {
        System.out.println(" num hyp " + hyp);
        System.out.println(sent.get(current));
      }

      x = 0;
      if (true) {
        // System.out.println("Exception "+current);
        hasHistory = false;
        //return;
      }
      if (!hasHistory) {
        histories = (getHistories(h));

        sum = 0;
        for (int tR = 0; tR < histories.length; tR++) {
          sum = sum + histories[tR];
        }
      }
      // Generate tags for current
      // Should implement dictionary later
      if ((allCount == 0)) {
        for (int y = 0; y < GlobalHolder.ySize; y++) {
          String tag = GlobalHolder.tags.getTag(y);
          if (GlobalHolder.tags.isClosed(tag)) {
            continue;
          }
          insertArray(current, hyp, x, y, hCurrent, pCurrent, tag, hasHistory, histories, sum);
        }
      } else { // the word is known and cut the search

        String[] tags1 = GlobalHolder.dict.getTags(sent.get(current));
        String[] tags = append(tags1, sent.get(current));

        for (int g = 0; g < tags.length; g++) {
          String tag = new String(tags[g]);
          int y = GlobalHolder.tags.getIndex(tag);
          //if(y==-1)
          // System.out.println(" y is " +y+" "+ tag);
          insertArray(current, hyp, x, y, hCurrent, pCurrent, tag, hasHistory, histories, sum);
        }

      }//else

      GlobalHolder.collectionTaggers.baseToken -= size;
    }//hyp

  }


  public String[] append(String[] tags, String word) {
    if (doDeterministicTagExpansion) {
      return GlobalHolder.tags.deterministicallyExpandTags(tags, word);
    } else {
      return tags;
    }
  }


  void insertArray(int current, int hyp, int x, int y, String[][] hCurrent, double[] pCurrent, String tag, boolean hasHistory, double[] histories, double sum) {
    double p = 0;

    if (DBG) {
      System.out.println(" inserting tag " + tag + " for word " + sent.get(current));
    }

    if (hasHistory) {
      //p=(pBest[hyp]>0?prob.pcond(y,x)*pBest[hyp]:prob.pcond(y,x));
      p = prob.pcond(y, x) * pBest[hyp];
    } else {
      //p=(histories[y]/sum)*pBest[hyp]; * chqange back to *
      p = Math.log(histories[y]) - Math.log(sum) + pBest[hyp];
    }//else no history
    if (!(p == p)) {
      System.out.println(" p ia NaN " + pBest[hyp] + " " + p + " current is" + current + " tag " + tag);
      return;
    }
    if (p == 0) {
      System.out.println(" P is 0 inside InsertArray");
      //smooth them

      return;
    }
    // this is acrobatics to avoid putting equal hypotheses on the beam
    for (int i = 0; i < beamSize; i++) {
      //here is the comparison
      if (p == pCurrent[i]) {
        boolean isDifferent = false;
        for (int j = 0; j < current; j++) {
          if (!(hCurrent[i][j].equals(nBest[hyp][j]))) {
            isDifferent = true;
          }
        }
        if (!(hCurrent[i][current].equals(tag))) {
          isDifferent = true;
        }
        if (!isDifferent) {
          return; // do not put the same thing twice
        }
      }
    }

    if (p < pCurrent[beamSize - 1]) {
      if (DBG) {
        System.out.println(" for word " + current + " did not insert tag " + tag + " prob " + (Math.log(histories[y]) - Math.log(sum)) + " total " + p);

      }
      return;
    }

    int i = beamSize - 2;
    while (true) {
      if ((i == -1) || (p < pCurrent[i])) {
        //Put the element at i+1
        pCurrent[i + 1] = p;
        hCurrent[i + 1][current] = tag;
        if (DBG) {
          System.out.println(" added for word " + current + " tag " + tag + " prob " + (Math.log(histories[y]) - Math.log(sum)) + " total " + p + " at place " + (i + 1));
        }

        for (int j = 0; j < current; j++) {
          hCurrent[i + 1][j] = nBest[hyp][j];
        }
        return;
      }
      // p>=p i
      // Write the p and the string for i on place i+1
      pCurrent[i + 1] = pCurrent[i];
      for (int j = 0; j < current + 1; j++) {
        hCurrent[i + 1][j] = hCurrent[i][j];
      }
      i--;
    } // End while
  }


  /**
   * This scores the current assignment in PairsHolder at
   * current position h.current
   */
  public double[] getScores(History h) {
    String cWord = ExtractorFrames.cWord.extract(h);
    String[] tags = stringTagsAt(h.current - h.start + leftWindow());
    if (DBG) {
      System.out.println("got " + tags.length + " tags for " + cWord);
      System.out.print("ttags has " + GlobalHolder.tags.getSize() + " tags: ");
      for(int i = 0; i < GlobalHolder.tags.getSize(); i++) System.out.print("\"" + GlobalHolder.tags.getTag(i) + "\" ");
      System.out.println();
    }

    // String ctag = ExtractorFrames.cTag.extract(h); //the current tag
    double[] scores = new double[tags.length];
    double[] histories = getHistories(h);
    double sum = 0.0;
    for (int i = 0; i < histories.length; i++) {
      sum += histories[i];
    }
    double logsum = Math.log(sum);

    //sum=0;

    for (int j = 0; j < tags.length; j++) {
      //score the j-th tag
      String tag = tags[j];
      int tagindex = GlobalHolder.tags.getIndex(tag);
      scores[j] = Math.log(histories[tagindex]);
      sum += histories[tagindex];
    }

    for (int j = 0; j < tags.length; j++) {
      scores[j] -= logsum;
    }

    if (DBG) {
      System.out.println("For " + h.current);
      System.out.print("  History features: ");
      h.print();
      for (int j = 0; j < tags.length; j++) {
        System.out.println("  Score for tag " + tags[j] + " " + scores[j]);
      }
    }

    return scores;
  }


  /**
   * print out the unknown word feature values of the features in ExtractorFramesRare
   */
  public void printActivations(History h) {
    String word = ExtractorFrames.cWord.extract(h);
    //FeatureKey s = new FeatureKey();

    System.out.print("features for word " + word + "\n");
    Extractors ext = GlobalHolder.extractorsRare;
    for (int j = 0; j < ext.getSize(); j++) {
      String key = ext.extract(j, h);
      System.out.print(j + "\t" + ext.toString() + " value " + key + "\n");
    }
    System.out.println();
  }


  // This is the most cpu intensive method in the tagging runtime!
  // It should do less string manipulation.
  // Changed to only do the exponentiation once -- cdm
  public double[] getHistories(History h) {
    double[] arrLocal = new double[GlobalHolder.ySize];
    String word = ExtractorFrames.cWord.extract(h);

    if (DBG) {
      System.out.println("getting probs for word " + ExtractorFrames.cWord.extract(h));
      System.out.println("previous tag is " + ExtractorFrames.prevTag.extract(h));
      System.out.println("previous two tags are " + ExtractorFrames.prevTwoTags.extract(h));
      h.print();
    }

    int fAll = ((GlobalHolder.isRare(ExtractorFrames.cWord.extract(h))) ? GlobalHolder.extractors.getSize() + GlobalHolder.extractorsRare.getSize() : GlobalHolder.extractors.getSize());
    //int fAll=GlobalHolder.extractors.getSize()+GlobalHolder.extractorsRare.getSize();
    int fMain = GlobalHolder.extractors.getSize();

    // reuse one FeatureKey rather than allocating each time for speed
    FeatureKey s = new FeatureKey();
    for (int i = 0; i < GlobalHolder.ySize; i++) {
      double lambda = 0.0;  // was =1;
      String tag = GlobalHolder.tags.getTag(i);
      for (int kf = 0; kf < fMain; kf++) {
        s.set(kf, GlobalHolder.extractors.extract(kf, h), tag);
        int fNum = GlobalHolder.getNum(s);
        if (fNum > -1) {
          // lambda=lambda*Math.exp(prob.lambda[fNum]);
          lambda += prob.lambda[fNum];
          if (kf > 3) {
            //System.out.println(s.toString()+" "+prob.lambda[fNum]);
          }
        }
      } // end for

      for (int kf = fMain; kf < fAll; kf++) {
        s.set(kf, GlobalHolder.extractorsRare.extract(kf - fMain, h), tag);
        int fNum = GlobalHolder.getNum(s);
        if (fNum > -1) {
          // lambda=lambda*Math.exp(prob.lambda[fNum]);
          lambda += prob.lambda[fNum];
          if (kf > 19) {
            //System.out.println("feature "+kf+"lambda "+prob.lambda[fNum]+" word "+word+" "+s.toString());
          }
        }
      } // end for
      // arrLocal[i]=lambda;

      //if we are multiplying in by the generation prob, do it here

      arrLocal[i] = Math.exp(lambda);
      if (DBG) {
        if (GlobalHolder.dict.getCount(word, tag) > 0) {
          System.out.println(" score for " + tag + " " + arrLocal[i]);
        }
      }

    } // end for tags
    // double sum=0;   // [cdm -- these additions weren't doing anything!]
    boolean smooth = false;
    for (int i = 0; i < arrLocal.length; i++) {
      // sum+=arrLocal[i];
      if (arrLocal[i] == 0) {
        smooth = true;
      }
    }
    if (smooth) {
      for (int i = 0; i < arrLocal.length; i++) {
        arrLocal[i] += .0001;
      }
    }//if
    return arrLocal;
  }


  /* ------------------------

  private static int[] intersect(int[] arr1, int[] arr2) {
    int[] result;
    HashSet<Integer> s = new HashSet<Integer>();
    if (arr1 == null) {
      return null;
    }
    if (arr2 == null) {
      return null;
    }
    for (int i = 0; i < arr1.length; i++) {
      for (int j = 0; j < arr2.length; j++) {
        if (arr1[i] == arr2[j]) {
          s.add(Integer.valueOf(arr1[i]));
          break;
        }
      }
    }
    if (s.isEmpty()) {
      return null;
    }
    result = new int[s.size()];
    Integer[] res = s.toArray(new Integer[0]);
    for (int i = 0; i < s.size(); i++) {
      result[i] = res[i].intValue();
    }
    return result;
  }

  private static int[] unite(int[] arr1, int[] arr2) {
    int[] result;
    HashSet<Integer> s = new HashSet<Integer>();
    if (arr1 == null) {
      return arr2;
    }
    if (arr2 == null) {
      return arr1;
    }
    for (int i = 0; i < arr1.length; i++) {
      s.add(Integer.valueOf(arr1[i]));
    }

    for (int i = 0; i < arr2.length; i++) {
      s.add(Integer.valueOf(arr2[i]));
    }
    if (s.isEmpty()) {
      return null;
    }
    result = new int[s.size()];
    Integer[] res = s.toArray(new Integer[0]);
    for (int i = 0; i < s.size(); i++) {
      result[i] = res[i].intValue();
    }
    return result;
  }

  -------------------------- */


  public double[][][] getProbs() {
    return probabilities;
  }


  public void printProbs() {
    for (int tag = 0; tag < GlobalHolder.tags.getSize(); tag++) {
      System.out.print(GlobalHolder.tags.getTag(tag) + "\t");
    }
    System.out.println();
    for (int current = 0; current < size; current++) {
      for (int hyp = 0; hyp < beamSize; hyp++) {
        System.out.print(sent.get(current) + "\t");
        for (int tag = 0; tag < GlobalHolder.tags.getSize(); tag++) {
          System.out.print((int) (1000 * probabilities[current][hyp][tag]) / (float) 1000 + "\t");
        }
        System.out.println();
      }// hyp
    }//current
  }

  /**
   * This method should be called after the sentence has been tagged. For every unknown word, this method adds the 3 most probable tags to the dictionary uDict
   */

  public void addUnknown(Dictionary uDict) {
    for (int current = 0; current < this.size; current++) {
      if (!known(sent.get(current))) {
        for (int i = 0; i < 3; i++) {
          uDict.add(sent.get(current), nBest[i][current]);
        }
      }// if
    }// for
  }


  /**
   * This method should be called after the sentence has been tagged. For every unknown word, this method prints the 3 most probable tags to the file pfu
   */
  public void printUnknown(int numSent, PrintFile pfu) {
    writeProbs();
    for (int current = 0; current < this.size; current++) {
      if (!known(sent.get(current))) {
        String s1 = sent.get(current) + ":" + numSent;
        double[] probs = new double[3];
        String[] tags = new String[3];
        getTop3(current, probs, tags);
        for (int i = 0; i < 3; i++) {
          s1 += "\t" + tags[i] + " " + probs[i];
        }
        pfu.println(s1);
        boolean wrong = true;
        String correctTag = toNice(this.correctTags[current]);
        for (int i = 0; i < 3; i++) {
          if (correctTag.equals(tags[i])) {
            wrong = false;
            break;
          }//if
        }
        if (wrong) {
          numWrongUnknownT3++;
        }
      }// if
    }// for
  }


  /**
   * This method should be called after the sentence has been tagged. For every word token, this method prints the 3 most probable tags to the file pfu
   * except for
   */
  public void printTop(PrintFile pfu) {
    writeProbs();
    for (int current = 0; current < this.size; current++) {
      if (true) {
        String s1 = sent.get(current);
        double[] probs = new double[3];
        String[] tags = new String[3];
        getTop3(current, probs, tags);
        for (int i = 0; i < 3; i++) {
          s1 += "\t" + tags[i] + " " + probs[i];
        }
        pfu.println(s1);
        boolean wrong = true;
        String correctTag = toNice(this.correctTags[current]);
        for (int i = 0; i < 3; i++) {
          if (correctTag.equals(tags[i])) {
            wrong = false;
            break;
          }//if
        }
        if (wrong) {
          numWrongT3++;
        }
      }// if
    }// for
  }


  void getTop3(int current, double[] probs, String[] tags) {
    int[] topIds = new int[3];
    double[] probTags = probabilities[current][0];
    for (int j = 0; j < 3; j++) {
      double maxP = 0.0;
      for (int i = 0; i < probTags.length; i++) {
        if (probTags[i] > maxP) {
          maxP = probTags[i];
          topIds[j] = i;
        }
      }//for
      probs[j] = probTags[topIds[j]];
      probTags[topIds[j]] = 0;
    }//j
    double total = 0;
    for (int j = 0; j < 3; j++) {
      total += probs[j];
    }
    for (int j = 0; j < 3; j++) {
      probs[j] /= total;
    }
    for (int j = 0; j < 3; j++) {
      tags[j] = toNice(GlobalHolder.tags.getTag(topIds[j]));
    }
  }


  /*
   * Implementation of the TagScorer interface follows
   */


  public int length() {
    return size;
  }

  public int leftWindow() {
    return leftContext; //hard-code for now
  }

  public int rightWindow() {
    return rightContext; //hard code for now
  }


  public int[] getPossibleValues(int pos) {
    String[] arr1 = stringTagsAt(pos);
    int[] arr = new int[arr1.length];
    for (int i = 0; i < arr.length; i++) {
      arr[i] = GlobalHolder.tags.getIndex(arr1[i]);
    }

    return arr;
  }


  public String[] stringTagsAt(int pos) {

    String[] arr1;
    if ((pos < leftWindow()) || (pos >= size + leftWindow())) {
      arr1 = new String[1];
      arr1[0] = naTag;
      return arr1;
    }

    int count = GlobalHolder.dict.sum(sent.get(pos - leftWindow()));
    if (count == 0) {
      arr1 = GlobalHolder.tags.getOpenTags().toArray(new String[0]);
    } else {
      arr1 = GlobalHolder.dict.getTags(sent.get(pos - leftWindow()));
    }

    arr1 = append(arr1, sent.get(pos - leftWindow()));

    if (VERBOSE) {
      //System.out.println("for word " + sent.get(pos - leftWindow()) + " count " + count + " tags are " + Arrays.toString(arr1));
    }

    return arr1;
  }


  public double scoreOf(int[] tags, int pos) {

    double[] scores = scoresOf(tags, pos);
    double score = Double.NEGATIVE_INFINITY;

    int[] pv = getPossibleValues(pos);

    for (int i = 0; i < scores.length; i++) {
      if (pv[i] == tags[pos]) {
        score = scores[i];
      }
    }

    return score;
  }

  public double scoreOf(int[] sequence) {
    throw new UnsupportedOperationException();
  }


  public double[] scoresOf(int[] tags, int pos) {
    History h = new History();

    if (DBG) {
      System.out.println("length of tags is " + tags.length);
    }

    h.set(endSizePairs - size, endSizePairs - 1, endSizePairs - size + pos - leftWindow());

    setHistory(pos, h, tags);
    return getScores(h);
  }


  /**
   * Tags a test sentence.
   *
   * @param args A single argument giving the filename of the parameter files.
   *             This should be a complete filename for holder file, for example,
   *             <code>~/mine/wsj0-20.holder</code>, but the program will also use
   *             other files by adding a second extension onto this filename.  If
   *             none is provided, a default tagger in <code>/u/nlp/data</code> is used.
   */
  public static void main(String[] args) {
    GlobalHolder.init();
    TestSentence tS = new TestSentence();
    BufferedReader br;

    System.err.println("Type a line to debug, Ctrl-C to exit.");
    br = new BufferedReader(new InputStreamReader(System.in));

    try {
      for (String temp; (temp = br.readLine()) != null; ) {
        tS.dumpActivations(temp);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

}
